#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>

#define N 256

/* Gestore del segnale SIGCHLD */
void handler(int s)
{
	int status;
	while (waitpid(-1, &status, WNOHANG) > 0);
}

int main(int argc, char** argv)
{
	struct addrinfo hints, *res;
	int err, sd, ns, pid, pid2, on, piped[2];
	char categoria[N];

	/* Controllo argomenti */
	if(argc < 2){
		printf("Uso: ./contocorrente-server-concurrent <porta> \n");
		exit(1);
	}

	signal(SIGCHLD, handler);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags    = AI_PASSIVE;

	if ((err = getaddrinfo(NULL, argv[1], &hints, &res)) != 0) {
		fprintf(stderr, "Errore setup indirizzo bind: %s\n", gai_strerror(err));
		exit(2);
	}

	if ((sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0){
		perror("Errore in socket");
		exit(3);
	}

	on = 1;
	if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0){
		perror("setsockopt");
		exit(4);
	}

	if (bind(sd, res->ai_addr, res->ai_addrlen) < 0) {
		perror("Errore in bind");
		exit(5);
	}

	freeaddrinfo(res);

	if(listen(sd, SOMAXCONN) < 0){
		perror("listen");
		exit(6);
	}

	/* Attendo i client... */
	for(;;){
		printf("Server in ascolto...\n");

		if((ns = accept(sd, NULL, NULL)) < 0) {
			/* Ignoro gli errori di tipo interruzione da segnale. */
			if (errno == EINTR)
				continue;
			/* Gestisco tutte le altre tipologie di errore. */
			perror("accept");
			exit(7);
		}

		/* Generazione di un figlio */
		if((pid = fork())<0){
			perror("fork");
			exit(8);
		} 
		else if(pid == 0){

			/* figlio */

			/* Chiudo la socket passiva */
			close(sd);

			memset(categoria, 0, sizeof(categoria));
			if(read(ns, categoria, N-1)<0){
				perror("read categoria");
				exit(9);
			}
			
			/* https://linux.die.net/man/3/pipe 
			int pipe(int fildes[2]);
			The pipe() function shall create a pipe and place two file descriptors, 
			one each into the arguments fildes[0] and fildes[1], 
			that refer to the open file descriptions for the read and write ends of the pipe. 
			Their integer values shall be the two lowest available at the time of the pipe() call. 
			[...] Upon successful completion, 0 shall be returned; 
			otherwise, -1 shall be returned and errno set to indicate the error.
			*/
			if (pipe(piped) < 0) {
				perror("pipe");
				exit(11);
			}

			if((pid2 = fork())<0){
				perror("seconda fork");
				exit(10);
			} 
			else if(pid2 == 0){

				// Nipote N1
				
				// Chiudo i descrittori che non servono
				close(piped[0]);
				close(ns);

				// Redireziono stdout
				close(1);
				dup(piped[1]);
				close(piped[1]);

				// Eseguo la grep mandando i risultati al figlio
				execlp("grep", "grep", categoria, "conto_corrente.txt", (char*)0);
				perror("execlp grep");
				
				exit(11);
			}

			// Figlio

			// Chiudo descrittori non necessari
			close(piped[1]);

			// Ridireziono stdin
			close(0);
			dup(piped[0]);
			close(piped[0]);

			/* Ridireziono stdout */
			close(1);
			dup(ns);
			close(ns);
			
			/* http://man7.org/linux/man-pages/man1/sort.1.html 
			sort - sort lines of text files 
			-n, --numeric-sort: compare according to string numerical value 
			-r, --reverse: reverse the result of comparisons */
			execlp("sort", "sort", "-r", "-n", (char*)0);
			perror("execlp sort");
			exit(13);

		}
		// padre
		close(ns);
	}
	return 0;
}
